import greenfoot.*;  // （World、Actor、GreenfootImage、GreenfootSound、Greenfoot及MouseInfo）
import java.awt.Point;
import java.util.ArrayList;
import java.util.List;

/**
 * 象棋棋盤
 * 
 * @author 依瑪貓
 * @version 2013-03-17
 */
public class Board extends World {
    
    /** 紅方是不是在棋盤下半部。 */
    private boolean isRedAtBottom = true;
    
    /** 現在輪到紅方或黑方。 */
    private ChessSide currentSide = ChessSide.RED;
    
    /** 第幾步的記步數。 */
    private int turns = 1;
    
    /** 上次吃子後的步數。 */
    private int lastCapture = 0;
    
    /** 棋局記錄。 */
    private List<String> history = new ArrayList<String>();
    
    /** 棋局。 */
    private static final String[] games = {
        // 空棋盤，供擺放棋局使用。
        "9/9/9/9/9/9/9/9/9/9 r - - 0 1",
        // 開局：標準開局。
        "rnbakabnr/9/1c5c1/p1p1p1p1p/9/9/P1P1P1P1P/1C5C1/9/RNBAKABNR r - - 0 1",  
        // 開局： 1999 許銀川對聶衛平 棋王對棋聖 讓九子（雙仕雙相五兵）
        // 參見： http://www.xqbase.com/protocol/cchess_pgn.htm
        "rnbakabnr/9/1c5c1/p1p1p1p1p/9/9/9/1C5C1/9/RN2K2NR r - - 0 1",
        // 殘局：炮兵勝雙士象。
        // 參見： http://www.chesstrip.org/finalphase/finalA236.htm
        "2b6/3ka4/2PaC4/9/9/9/9/9/9/4K4 r - - 0 1",
        // 殘局：小紅馬。
        // 參見： http://lkcsschessclub.blogspot.tw/2011/01/blog-post.html
        "1C1k5/2P2n3/5n3/9/9/9/7NR/9/3r5/4Kc1r1 r - - 0 1",
        // 殘局：橘中秘 殘局第四局 借炮使傌
        "2bak4/4a4/1C2b4/4pNp2/9/9/9/2n1BCN2/3rAp3/1cBAK4 r - - 0 1"
        };
        
    
    /**
     * 建立新的象棋棋盤。
     * 
     */
    public Board() {
        // 建立9x10方格的新場景，方格大小為1x1像素。
        super(9, 10, 50); 
        isRedAtBottom = true;
        currentSide = ChessSide.RED;
        addObject(new CurrentSide(currentSide), 4, 4);
        initializeGame(1);
        if (!isRedAtBottom) {
            GreenfootImage background = getBackground();
            background.rotate(180);
        }
    }
    
    /**
     * 每一回合執行的動作。
     * 
     */
    public void act() {
        List<Result> results = getObjects(Result.class);
        if (results.size() > 0) {
            Greenfoot.stop();
        }
    }
    
    /**
     * 紅方是不是在棋盤下半部。
     * 
     * @return 紅方在棋盤下半部時傳回 true ，否則傳回 false 。
     */
    public boolean isRedAtBottom() {
        return isRedAtBottom;
    }
    
    /**
     * 取得某一格的棋子。
     * 
     * @param x 那一格的 X 座標
     * @param y 那一格的 Y 座標
     * @return 那一格有棋子時傳回該棋子，否則傳回 null 。
     */
    public Piece getPieceAt(int x, int y) {
        List<Piece> pieces = getObjectsAt(x, y, Piece.class);
        if (pieces.size() > 0) {
            return pieces.get(0);
        }
        return null;
    }
    
    /**
     * 某一格有沒有棋子。
     * 
     * @param x 那一格的 X 座標
     * @param y 那一格的 Y 座標
     * @return 那一格有棋子時傳回 true ，否則傳回 false 。
     */
    public boolean isOccupiedAt(int x, int y) {
        List<Piece> pieces = getObjectsAt(x, y, Piece.class);
        if (pieces.size() > 0) {
            return true;
        }
        return false;
    }
    
    /**
     * 取得現在輪到的陣營，紅方或黑方。
     * 
     * @return 現在輪到的陣營，紅方或黑方。
     */
    public ChessSide getCurrentSide() {
        return currentSide;
    }
    
    /**
     * 換另一方。
     * 
     */
    public void switchSide() {
        checkGameOver();
        switch (currentSide) {
        case RED:
            currentSide = ChessSide.BLACK;
            break;
        case BLACK:
            currentSide = ChessSide.RED;
            break;
        }
        turns++;
        lastCapture++;
        history.add(getFen());
        List<CurrentSide> sides = getObjects(CurrentSide.class);
        sides.get(0).switchSide();
    }
    
    /**
     * 棋盤上下翻轉。
     * 
     */
    public void rotateBoard() {
        GreenfootImage background = getBackground();
        List<Piece> pieces = getObjects(Piece.class);
        
        isRedAtBottom = !isRedAtBottom;
        background.rotate(180);
        for (Piece piece : pieces) {
            GreenfootImage image = piece.getImage();
            image.rotate(180);
            piece.setLocation(8 - piece.getX(), 9 - piece.getY());
        }
    }
    
    /**
     * 吃子後的步數歸零。吃子後呼叫。
     * 
     */
    public void resetLastCapture() {
        lastCapture = 0;
    }
    
    /**
     * 擺放棋局。
     * 
     * @param num 第幾個棋局： 0 是空棋盤，供擺設棋局用； 1 是標準開局
     */
    public void initializeGame(int num) {
        setUpByFen(games[num]);
        history.clear();
        history.add(games[num]);
    }
    
    /**
     * 顯示棋局的 FEN 表示法。見：
     * http://www.xqbase.com/protocol/pgnfen2.htm
     * 
     */
    public void printFen() {
        System.out.println(getFen());
    }
    
    /**
     * 取得棋局的 FEN 表示法。見：
     * http://www.xqbase.com/protocol/pgnfen2.htm
     * 
     * @return 棋局的 FEN 表示法
     */
    private String getFen() {
        String fen = "";
        int spaces = 0;
        
        for (int y = 0; y <= 9; y++) {
            for (int x = 0; x <= 8; x++) {
                Piece piece = getPieceAt(x, y);
                if (piece == null) {
                    spaces++;
                } else {
                    if (spaces > 0) {
                        fen += spaces;
                        spaces = 0;
                    }
                    if (piece instanceof General) {
                        if (piece.getSide() == ChessSide.RED) {
                            fen += "K";
                        } else {
                            fen += "k";
                        }
                    } else if (piece instanceof Advisor) {
                        if (piece.getSide() == ChessSide.RED) {
                            fen += "A";
                        } else {
                            fen += "a";
                        }
                    } else if (piece instanceof Elephant) {
                        if (piece.getSide() == ChessSide.RED) {
                            fen += "B";
                        } else {
                            fen += "b";
                        }
                    } else if (piece instanceof Chariot) {
                        if (piece.getSide() == ChessSide.RED) {
                            fen += "R";
                        } else {
                            fen += "r";
                        }
                    } else if (piece instanceof Horse) {
                        if (piece.getSide() == ChessSide.RED) {
                            fen += "N";
                        } else {
                            fen += "n";
                        }
                    } else if (piece instanceof Cannon) {
                        if (piece.getSide() == ChessSide.RED) {
                            fen += "C";
                        } else {
                            fen += "c";
                        }
                    } else if (piece instanceof Soldier) {
                        if (piece.getSide() == ChessSide.RED) {
                            fen += "P";
                        } else {
                            fen += "p";
                        }
                    }
                }
            }
            if (spaces > 0) {
                fen += spaces;
                spaces = 0;
            }
            if (y < 9) {
                fen += "/";
            }
        }
        
        fen += " ";
        if (currentSide == ChessSide.RED) {
            fen += "r";
        } else {
            fen += "b";
        }
        fen += " - - " + lastCapture + " " + turns;
        return fen;
    }
    
    /**
     * 依 FEN 表示法還原棋局。見：
     * http://www.xqbase.com/protocol/pgnfen2.htm
     * 
     * @param fen 棋局的 FEN 表示法
     */
    private void setUpByFen(String fen) {
        String[] parts = new String[0];
        String[] rows = new String[0];
        int x = 0, y = 0;
        int x0 = -1, y0 = -1, step = 0;
        
        parts = fen.trim().split(" ");
        rows = parts[0].split("/");
        
        if (isRedAtBottom) {
            x0 = 0;
            y0 = 0;
            step = 1;
        } else {
            x0 = 8;
            y0 = 9;
            step = -1;
        }
        
        removeObjects(getObjects(Piece.class));
        y = 0;
        for (String row : rows) {
            x = 0;
            for (int i = 0; i < row.length(); i++) {
                char c = row.charAt(i);
                switch (c) {
                case 'K':
                    addObject(new General(ChessSide.RED), x0 + step * x, y0 + step * y);
                    x++;
                    break;
                case 'k':
                    addObject(new General(ChessSide.BLACK), x0 + step * x, y0 + step * y);
                    x++;
                    break;
                case 'A':
                    addObject(new Advisor(ChessSide.RED), x0 + step * x, y0 + step * y);
                    x++;
                    break;
                case 'a':
                    addObject(new Advisor(ChessSide.BLACK), x0 + step * x, y0 + step * y);
                    x++;
                    break;
                case 'B':
                    addObject(new Elephant(ChessSide.RED), x0 + step * x, y0 + step * y);
                    x++;
                    break;
                case 'b':
                    addObject(new Elephant(ChessSide.BLACK), x0 + step * x, y0 + step * y);
                    x++;
                    break;
                case 'R':
                    addObject(new Chariot(ChessSide.RED), x0 + step * x, y0 + step * y);
                    x++;
                    break;
                case 'r':
                    addObject(new Chariot(ChessSide.BLACK), x0 + step * x, y0 + step * y);
                    x++;
                    break;
                case 'N':
                    addObject(new Horse(ChessSide.RED), x0 + step * x, y0 + step * y);
                    x++;
                    break;
                case 'n':
                    addObject(new Horse(ChessSide.BLACK), x0 + step * x, y0 + step * y);
                    x++;
                    break;
                case 'C':
                    addObject(new Cannon(ChessSide.RED), x0 + step * x, y0 + step * y);
                    x++;
                    break;
                case 'c':
                    addObject(new Cannon(ChessSide.BLACK), x0 + step * x, y0 + step * y);
                    x++;
                    break;
                case 'P':
                    addObject(new Soldier(ChessSide.RED), x0 + step * x, y0 + step * y);
                    x++;
                    break;
                case 'p':
                    addObject(new Soldier(ChessSide.BLACK), x0 + step * x, y0 + step * y);
                    x++;
                    break;
                case '1':
                case '2':
                case '3':
                case '4':
                case '5':
                case '6':
                case '7':
                case '8':
                case '9':
                    int spaces = c - '0';
                    x += spaces;
                    break;
                }
            }
            y++;
        }
        
        if (parts[1].charAt(0) == 'b') {
            currentSide = ChessSide.BLACK;
        } else {
            currentSide = ChessSide.RED;
        }
        lastCapture = Integer.parseInt(parts[4]);
        turns = Integer.parseInt(parts[5]);
    }
    
    /**
     * 檢查是否遊戲結束。
     * 
     */
    private void checkGameOver() {
        List<General> generals = getObjects(General.class);
        if (generals.size() != 2) {
            Greenfoot.stop();
        }
        List<Piece> pieces = getObjects(Piece.class);
        boolean isRedAvailable = false, isBlackAvailable = false;
        for (Piece piece : pieces) {
            switch (piece.getSide()) {
            case RED:
                if (!isRedAvailable) {
                    if (piece.getCapturables().size() > 0 || piece.getMovables().size() > 0) {
                        isRedAvailable = true;
                    }
                }
                break;
            case BLACK:
                if (!isBlackAvailable) {
                    if (piece.getCapturables().size() > 0 || piece.getMovables().size() > 0) {
                        isBlackAvailable = true;
                    }
                }
                break;
            }
        }
        if (!isRedAvailable) {
            addObject(new Result(ChessSide.BLACK), 4, 4);
            Greenfoot.stop();
        } else if (!isBlackAvailable) {
            addObject(new Result(ChessSide.RED), 4, 4);
            Greenfoot.stop();
        }
    }
}