import greenfoot.*;  // （World、Actor、GreenfootImage、GreenfootSound、Greenfoot及MouseInfo）
import java.awt.Point;
import java.util.ArrayList;
import java.util.List;

/**
 * 象棋的炮／包。
 * 
 * @author 依瑪貓
 * @version 2013-03-17
 */
public class Cannon extends Piece {
    
    /**
     * 建立新的炮／包。
     * 
     * @param chessSide 棋子所屬的陣營，紅方或黑方。
     */
    public Cannon(ChessSide chessSide) {
        super(chessSide);
    }
    
    /**
     * 能不能吃掉另一顆棋子。
     * 
     * @param piece 另一顆棋子
     * @return 如果可以吃掉那顆棋子，就傳回 true ，否則傳回 false 。
     */
    public boolean isCapturable(Piece another) {
        Board board = (Board) getWorld();
        
        // 不能吃同一方的棋子。
        if (getSide() == another.getSide()) {
            return false;
        }
        
        // 同一橫行。
        if (getX() == another.getX()) {
            int y0 = -1, y1 = -1, between = 0;
            if (getY() < another.getY()) {
                y0 = getY();
                y1 = another.getY();
            } else {
                y0 = another.getY();
                y1 = getY();
            }
            for (int y = y0 + 1; y < y1; y++) {
                if (board.isOccupiedAt(getX(), y)) {
                    between++;
                }
            }
            if (between == 1) {
                return true;
            } else {
                return false;
            }
        }
        
        // 同一縱路。
        if (getY() == another.getY()) {
            int x0 = -1, x1 = -1, between = 0;
            if (getX() < another.getX()) {
                x0 = getX();
                x1 = another.getX();
            } else {
                x0 = another.getX();
                x1 = getX();
            }
            for (int x = x0 + 1; x < x1; x++) {
                if (board.isOccupiedAt(x, getY())) {
                    between++;
                }
            }
            if (between == 1) {
                return true;
            } else {
                return false;
            }
        }
        
        return false;
    }
    
    /**
     * 能不能移到另一格。
     * 
     * @param targetX 另一格的 X 座標
     * @param targetY 另一格的 Y 座標
     * @return 如果可以移到那一格，就傳回 true ，否則傳回 false 。
     */
    public boolean isMovable(int targetX, int targetY) {
        Board board = (Board) getWorld();
        
        // 只能移到空格。
        if (board.isOccupiedAt(targetX, targetY)) {
            return false;
        }
        
        // 同一橫行。
        if (targetX == getX()) {
            int y0 = -1, y1 = -1;
            if (getY() < targetY) {
                y0 = getY();
                y1 = targetY;
            } else {
                y0 = targetY;
                y1 = getY();
            }
            for (int y = y0 + 1; y < y1; y++) {
                if (board.isOccupiedAt(getX(), y)) {
                    return false;
                }
            }
            return true;
        }
        
        // 同一縱路。
        if (targetY == getY()) {
            int x0 = -1, x1 = -1;
            if (getX() < targetX) {
                x0 = getX();
                x1 = targetX;
            } else {
                x0 = targetX;
                x1 = getX();
            }
            for (int x = x0 + 1; x < x1; x++) {
                if (board.isOccupiedAt(x, getY())) {
                    return false;
                }
            }
            return true;
        }
        
        return false;
    }
    
    /**
     * 取得可以吃的棋子。
     * 
     * @return 可以吃的棋子。
     */
    public List<Piece> getCapturables() {
        Board board = (Board) getWorld();
        List<Piece> capturables = new ArrayList<Piece>();
        
        for (int x = getX() - 1; x >= 0; x--) {
            if (board.isOccupiedAt(x, getY())) {
                for (x = x - 1; x >= 0; x--) {
                    Piece piece = board.getPieceAt(x, getY());
                    if (piece != null) {
                        if (getSide() != piece.getSide()) {
                            capturables.add(piece);
                        }
                        break;
                    }
                }
                break;
            }
        }
        for (int x = getX() + 1; x < 9; x++) {
            if (board.isOccupiedAt(x, getY())) {
                for (x = x + 1; x < 9; x++) {
                    Piece piece = board.getPieceAt(x, getY());
                    if (piece != null) {
                        if (getSide() != piece.getSide()) {
                            capturables.add(piece);
                        }
                        break;
                    }
                }
                break;
            }
        }
        for (int y = getY() - 1; y >= 0; y--) {
            if (board.isOccupiedAt(getX(), y)) {
                for (y = y - 1; y >= 0; y--) {
                    Piece piece = board.getPieceAt(getX(), y);
                    if (piece != null) {
                        if (getSide() != piece.getSide()) {
                            capturables.add(piece);
                        }
                        break;
                    }
                }
                break;
            }
        }
        for (int y = getY() + 1; y < 10; y++) {
            if (board.isOccupiedAt(getX(), y)) {
                for (y = y + 1; y < 10; y++) {
                    Piece piece = board.getPieceAt(getX(), y);
                    if (piece != null) {
                        if (getSide() != piece.getSide()) {
                            capturables.add(piece);
                        }
                        break;
                    }
                }
                break;
            }
        }
        
        // 刪掉會被將軍，或無法解套將軍的吃法。
        removeCheckmatedEats(capturables);
        return capturables;
    }
    
    /**
     * 取得可以走的空格。
     * 
     * @return 可以走的空格。
     */
    public List<Point> getMovables() {
        Board board = (Board) getWorld();
        List<Point> movables = new ArrayList<Point>();
        
        for (int x = getX() - 1; x >= 0; x--) {
            if (board.isOccupiedAt(x, getY())) {
                break;
            }
            movables.add(new Point(x, getY()));
        }
        for (int x = getX() + 1; x < 9; x++) {
            if (board.isOccupiedAt(x, getY())) {
                break;
            }
            movables.add(new Point(x, getY()));
        }
        for (int y = getY() - 1; y >= 0; y--) {
            if (board.isOccupiedAt(getX(), y)) {
                break;
            }
            movables.add(new Point(getX(), y));
        }
        for (int y = getY() + 1; y < 10; y++) {
            if (board.isOccupiedAt(getX(), y)) {
                break;
            }
            movables.add(new Point(getX(), y));
        }
        
        // 刪掉會被將軍，或無法解套將軍的走法。
        removeCheckmatedMoves(movables);
        return movables;
    }
}