import greenfoot.*;  // （World、Actor、GreenfootImage、GreenfootSound、Greenfoot及MouseInfo）
import java.awt.Point;
import java.util.ArrayList;
import java.util.List;

/**
 * 象棋的相／象。
 * 
 * @author 依瑪貓
 * @version 2013-03-17
 */
public class Elephant extends Piece {
    
    /**
     * 建立新的相／象。
     * 
     * @param chessSide 棋子所屬的陣營，紅方或黑方。
     */
    public Elephant(ChessSide chessSide) {
        super(chessSide);
    }
    
    /**
     * 能不能吃掉另一顆棋子。
     * 
     * @param piece 另一顆棋子
     * @return 如果可以吃掉那顆棋子，就傳回 true ，否則傳回 false 。
     */
public boolean isCapturable(Piece another) {
        // 棋子只能放在符合規則的位置上。
        if (!isAtLegalPosition()) {
            return false;
        }
        // 不能吃同一方的棋子。
        if (getSide() == another.getSide()) {
            return false;
        }
        // 檢查另一顆棋子是不是在移動範圍內。
        return isInMovingRange(another.getX(), another.getY());
    }
    
    /**
     * 能不能移到另一格。
     * 
     * @param targetX 另一格的 X 座標
     * @param targetY 另一格的 Y 座標
     * @return 如果可以移到那一格，就傳回 true ，否則傳回 false 。
     */
    public boolean isMovable(int targetX, int targetY) {
        Board board = (Board) getWorld();
        
        // 棋子只能放在符合規則的位置上。
        if (!isAtLegalPosition()) {
            return false;
        }
        // 只能移到空格。
        if (board.isOccupiedAt(targetX, targetY)) {
            return false;
        }
        // 檢查目標格是不是在移動範圍內。
        return isInMovingRange(targetX, targetY);
    }
    
    /**
     * 取得可以吃的棋子。
     * 
     * @return 可以吃的棋子。
     */
    public List<Piece> getCapturables() {
        Board board = (Board) getWorld();
        List<Piece> capturables = new ArrayList<Piece>();
        int y0 = -1, y1 = -1;
        
        // 棋子只能放在符合規則的位置上。
        if (!isAtLegalPosition()) {
            return capturables;
        }
        
        if (isBottomSide()) {
            y0 = 6;
            y1 = 8;
        } else {
            y0 = 1;
            y1 = 3;
        }
        
        // 左方。
        if (getX() > 1) {
            // 左上方。
            if (getY() > y0) {
                // 沒有拐象眼。
                if (!board.isOccupiedAt(getX() - 1, getY() - 1)) {
                    Piece piece = board.getPieceAt(getX() - 2, getY() - 2);
                    if (piece != null) {
                        if (getSide() != piece.getSide()) {
                            capturables.add(piece);
                        }
                    }
                }
            }
            // 左下方。
            if (getY() < y1) {
                // 沒有拐象眼。
                if (!board.isOccupiedAt(getX() - 1, getY() + 1)) {
                    Piece piece = board.getPieceAt(getX() - 2, getY() + 2);
                    if (piece != null) {
                        if (getSide() != piece.getSide()) {
                            capturables.add(piece);
                        }
                    }
                }
            }
        }
        // 右方。
        if (getX() < 7) {
            // 右上方。
            if (getY() > y0) {
                // 沒有拐象眼。
                if (!board.isOccupiedAt(getX() + 1, getY() - 1)) {
                    Piece piece = board.getPieceAt(getX() + 2, getY() - 2);
                    if (piece != null) {
                        if (getSide() != piece.getSide()) {
                            capturables.add(piece);
                        }
                    }
                }
            }
            // 右下方。
            if (getY() < y1) {
                // 沒有拐象眼。
                if (!board.isOccupiedAt(getX() + 1, getY() + 1)) {
                    Piece piece = board.getPieceAt(getX() + 2, getY() + 2);
                    if (piece != null) {
                        if (getSide() != piece.getSide()) {
                            capturables.add(piece);
                        }
                    }
                }
            }
        }
        
        // 刪掉會被將軍，或無法解套將軍的吃法。
        removeCheckmatedEats(capturables);
        return capturables;
    }
    
    /**
     * 取得可以走的空格。
     * 
     * @return 可以走的空格。
     */
    public List<Point> getMovables() {
        Board board = (Board) getWorld();
        List<Point> movables = new ArrayList<Point>();
        int y0 = -1, y1 = -1;
        
        // 棋子只能放在符合規則的位置上。
        if (!isAtLegalPosition()) {
            return movables;
        }
        
        if (isBottomSide()) {
            y0 = 6;
            y1 = 8;
        } else {
            y0 = 1;
            y1 = 3;
        }
        
        // 左方。
        if (getX() > 1) {
            // 左上方。
            if (getY() > y0) {
                // 沒有拐象眼。
                if (!board.isOccupiedAt(getX() - 1, getY() - 1)) {
                    if (!board.isOccupiedAt(getX() - 2, getY() - 2)) {
                        movables.add(new Point(getX() - 2, getY() - 2));
                    }
                }
            }
            // 左下方。
            if (getY() < y1) {
                // 沒有拐象眼。
                if (!board.isOccupiedAt(getX() - 1, getY() + 1)) {
                    if (!board.isOccupiedAt(getX() - 2, getY() + 2)) {
                        movables.add(new Point(getX() - 2, getY() + 2));
                    }
                }
            }
        }
        // 右方。
        if (getX() < 7) {
            // 右上方。
            if (getY() > y0) {
                // 沒有拐象眼。
                if (!board.isOccupiedAt(getX() + 1, getY() - 1)) {
                    if (!board.isOccupiedAt(getX() + 2, getY() - 2)) {
                        movables.add(new Point(getX() + 2, getY() - 2));
                    }
                }
            }
            // 右下方。
            if (getY() < y1) {
                // 沒有拐象眼。
                if (!board.isOccupiedAt(getX() + 1, getY() + 1)) {
                    if (!board.isOccupiedAt(getX() + 2, getY() + 2)) {
                        movables.add(new Point(getX() + 2, getY() + 2));
                    }
                }
            }
        }
        
        // 刪掉會被將軍，或無法解套將軍的走法。
        removeCheckmatedMoves(movables);
        return movables;
    }
    
    /**
     * 棋子所在是不是符合規則的位置上。
     * 
     * @return 如果棋子在符合規則的位置上，就傳回 true ，否則傳回 false 。
     */
    public boolean isAtLegalPosition() {
        int y0 = -1, y1 = -1, dy = 0;
        
        if (isBottomSide()) {
            y0 = 5;
            y1 = 9;
            dy = 9 - getY();
        } else {
            y0 = 0;
            y1 = 4;
            dy = getY();
        }
        
        if (getX() < 0 || getX() > 8 || getY() < y0 || getY() > y1) {
            return false;
        }
        if ((getX() + dy) % 4 != 2) {
            return false;
        }
        return true;
    }
    
    /**
     * 另一格是不是在移動範圍內。
     * 
     * @param targetX 另一格的 X 座標
     * @param targetY 另一格的 Y 座標
     * @return 如果另一格在移動範圍內，就傳回 true ，否則傳回 false 。
     */
    private boolean isInMovingRange(int targetX, int targetY) {
        Board board = (Board) getWorld();
        int dx = targetX - getX();
        int dy = targetY - getY();
        
        if (isBottomSide()) {
            if (targetY < 5) {
                return false;
            }
        } else {
            if (targetY > 4) {
                return false;
            }
        }
        
        if (dx * dx + dy * dy != 8) {
            return false;
        }
        // 拐象眼。
        if (board.isOccupiedAt(getX() + dx / 2, getY() + dy / 2)) {
            return false;
        }
        return true;
    }
}