import greenfoot.*;  // （World、Actor、GreenfootImage、GreenfootSound、Greenfoot及MouseInfo）
import java.awt.Point;
import java.util.ArrayList;
import java.util.List;

/**
 * 象棋的帥／將。
 * 
 * @author 依瑪貓
 * @version 2013-03-17
 */
public class General extends Piece {
    
    /**
     * 建立新的帥／將。
     * 
     * @param chessSide 棋子所屬的陣營，紅方或黑方。
     */
    public General(ChessSide chessSide) {
        super(chessSide);
    }
    
    /**
     * 能不能吃掉另一顆棋子。
     * 
     * @param piece 另一顆棋子
     * @return 如果可以吃掉那顆棋子，就傳回 true ，否則傳回 false 。
     */
    public boolean isCapturable(Piece another) {
        // 棋子只能放在符合規則的位置上。
        if (!isAtLegalPosition()) {
            return false;
        }
        // 不能吃同一方的棋子。
        if (getSide() == another.getSide()) {
            return false;
        }
        // 王不見王。
        if (isSeeingAnotherGeneral(another)) {
            return true;
        }
        // 檢查另一顆棋子是不是在移動範圍內。
        return isInMovingRange(another.getX(), another.getY());
    }
    
    /**
     * 能不能移到另一格。
     * 
     * @param targetX 另一格的 X 座標
     * @param targetY 另一格的 Y 座標
     * @return 如果可以移到那一格，就傳回 true ，否則傳回 false 。
     */
    public boolean isMovable(int targetX, int targetY) {
        Board board = (Board) getWorld();
        
        // 棋子只能放在符合規則的位置上。
        if (!isAtLegalPosition()) {
            return false;
        }
        // 只能移到空格。
        if (board.isOccupiedAt(targetX, targetY)) {
            return false;
        }
        // 檢查目標格是不是在移動範圍內。
        return isInMovingRange(targetX, targetY);
    }
    
    /**
     * 取得可以吃的棋子。
     * 
     * @return 可以吃的棋子。
     */
    public List<Piece> getCapturables() {
        Board board = (Board) getWorld();
        List<Piece> capturables = new ArrayList<Piece>();
        int y0 = -1, y1 = -1;
        
        // 棋子只能放在符合規則的位置上。
        if (!isAtLegalPosition()) {
            return capturables;
        }
        
        if (isBottomSide()) {
            y0 = 7;
            y1 = 9;
        } else {
            y0 = 0;
            y1 = 2;
        }
        
        // 左邊。
        if (getX() > 3) {
            Piece piece = board.getPieceAt(getX() - 1, getY());
            if (piece != null) {
                if (getSide() != piece.getSide()) {
                    capturables.add(piece);
                }
            }
        }
        // 右邊。
        if (getX() < 5) {
            Piece piece = board.getPieceAt(getX() + 1, getY());
            if (piece != null) {
                if (getSide() != piece.getSide()) {
                    capturables.add(piece);
                }
            }
        }
        // 上面。
        if (getY() > y0) {
            Piece piece = board.getPieceAt(getX(), getY() - 1);
            if (piece != null) {
                if (getSide() != piece.getSide()) {
                    capturables.add(piece);
                }
            }
        }
        // 下面。
        if (getY() < y1) {
            Piece piece = board.getPieceAt(getX(), getY() + 1);
            if (piece != null) {
                if (getSide() != piece.getSide()) {
                    capturables.add(piece);
                }
            }
        }
        // 王不見王。
        if (isBottomSide()) {
            for (int y = getY() - 1; y >= 0; y--) {
                Piece piece = board.getPieceAt(getX(), y);
                if (piece != null) {
                    if (piece instanceof General && getSide() != piece.getSide()) {
                        capturables.add(piece);
                    }
                    break;
                }
            }
        } else {
            for (int y = getY() + 1; y < 10; y++) {
                Piece piece = board.getPieceAt(getX(), y);
                if (piece != null) {
                    if (piece instanceof General && getSide() != piece.getSide()) {
                        capturables.add(piece);
                    }
                    break;
                }
            }
        }
        
        // 刪掉會被將軍，或無法解套將軍的吃法。
        removeCheckmatedEats(capturables);
        return capturables;
    }
    
    /**
     * 取得可以走的空格。
     * 
     * @return 可以走的空格。
     */
    public List<Point> getMovables() {
        Board board = (Board) getWorld();
        List<Point> movables = new ArrayList<Point>();
        int y0 = -1, y1 = -1;
        
        // 棋子只能放在符合規則的位置上。
        if (!isAtLegalPosition()) {
            return movables;
        }
        
        if (isBottomSide()) {
            y0 = 7;
            y1 = 9;
        } else {
            y0 = 0;
            y1 = 2;
        }
        
        // 左邊。
        if (getX() > 3) {
            if (!board.isOccupiedAt(getX() - 1, getY())) {
                movables.add(new Point(getX() - 1, getY()));
            }
        }
        // 右邊。
        if (getX() < 5) {
            if (!board.isOccupiedAt(getX() + 1, getY())) {
                movables.add(new Point(getX() + 1, getY()));
            }
        }
        // 上面。
        if (getY() > y0) {
            if (!board.isOccupiedAt(getX(), getY() - 1)) {
                movables.add(new Point(getX(), getY() - 1));
            }
        }
        // 下面。
        if (getY() < y1) {
            if (!board.isOccupiedAt(getX(), getY() + 1)) {
                movables.add(new Point(getX(), getY() + 1));
            }
        }
        
        // 刪掉會被將軍，或無法解套將軍的走法。
        removeCheckmatedMoves(movables);
        return movables;
    }
    
    /**
     * 棋子所在是不是符合規則的位置上。
     * 
     * @return 如果棋子在符合規則的位置上，就傳回 true ，否則傳回 false 。
     */
    public boolean isAtLegalPosition() {
        int y0 = -1, y1 = -1, dy = 0;
        
        if (isBottomSide()) {
            y0 = 7;
            y1 = 9;
            dy = 9 - getY();
        } else {
            y0 = 0;
            y1 = 2;
            dy = getY();
        }
        
        if (getX() < 3 || getX() > 5 || getY() < y0 || getY() > y1) {
            return false;
        }
        return true;
    }
    
    /**
     * 是不是被將軍了。
     * 
     * @return 如果被將軍了，就傳回 true ，否則傳回 false 。
     */
    public boolean isCheckmated() {
        Board board = (Board) getWorld();
        List<Piece> pieces = board.getObjects(Piece.class);
        for (Piece piece : pieces) {
            if (piece.isCapturable(this)) {
                return true;
            }
        }
        return false;
    }
    
    /**
     * 是不是王見王。
     * 
     * @param piece 另一顆棋子
     * @return 如果是王見王，就傳回 true ，否則傳回 false 。
     */
    private boolean isSeeingAnotherGeneral(Piece another) {
        Board board = (Board) getWorld();
        int y0 = -1, y1 = -1;
        
        if (!(another instanceof General)) {
            return false;
        }
        if (getX() != another.getX()) {
            return false;
        }
        if (getY() < another.getY()) {
            y0 = getY();
            y1 = another.getY();
        } else {
            y0 = another.getY();
            y1 = getY();
        }
        for (int y = y0 + 1; y < y1; y++) {
            if (board.isOccupiedAt(getX(), y)) {
                return false;
            }
        }
        return true;
    }
    
    /**
     * 另一格是不是在移動範圍內。
     * 
     * @param targetX 另一格的 X 座標
     * @param targetY 另一格的 Y 座標
     * @return 如果另一格在移動範圍內，就傳回 true ，否則傳回 false 。
     */
    private boolean isInMovingRange(int targetX, int targetY) {
        int dx = targetX - getX();
        int dy = targetY - getY();
        
        if (targetX < 3 || targetX > 5) {
            return false;
        }
        if (isBottomSide()) {
            if (targetY < 7) {
                return false;
            }
        } else {
            if (targetY > 2) {
                return false;
            }
        }
        
        if (dx * dx + dy * dy != 1) {
            return false;
        }
        return true;
    }
}