import greenfoot.*;  // （World、Actor、GreenfootImage、GreenfootSound、Greenfoot及MouseInfo）
import java.awt.Point;
import java.util.ArrayList;
import java.util.List;

/**
 * 象棋的傌／馬。
 * 
 * @author 依瑪貓
 * @version 2013-03-17
 */
public class Horse extends Piece {
    
    /**
     * 建立新的傌／馬。
     * 
     * @param chessSide 棋子所屬的陣營，紅方或黑方。
     */
    public Horse(ChessSide chessSide) {
        super(chessSide);
    }
    
    /**
     * 能不能吃掉另一顆棋子。
     * 
     * @param piece 另一顆棋子
     * @return 如果可以吃掉那顆棋子，就傳回 true ，否則傳回 false 。
     */
    public boolean isCapturable(Piece another) {
        // 不能吃同一方的棋子。
        if (getSide() == another.getSide()) {
            return false;
        }
        // 檢查另一顆棋子是不是在移動範圍內。
        return isInMovingRange(another.getX(), another.getY());
    }
    
    /**
     * 能不能移到另一格。
     * 
     * @param targetX 另一格的 X 座標
     * @param targetY 另一格的 Y 座標
     * @return 如果可以移到那一格，就傳回 true ，否則傳回 false 。
     */
    public boolean isMovable(int targetX, int targetY) {
        Board board = (Board) getWorld();
        
        // 只能移到空格。
        if (board.isOccupiedAt(targetX, targetY)) {
            return false;
        }
        // 檢查目標格是不是在移動範圍內。
        return isInMovingRange(targetX, targetY);
    }
    
    /**
     * 取得可以吃的棋子。
     * 
     * @return 可以吃的棋子。
     */
    public List<Piece> getCapturables() {
        Board board = (Board) getWorld();
        List<Piece> capturables = new ArrayList<Piece>();
        
        // 左邊。
        if (getX() > 1) {
            if (!board.isOccupiedAt(getX() - 1, getY())) {
                if (getY() > 0) {
                    Piece piece = board.getPieceAt(getX() - 2, getY() - 1);
                    if (piece != null) {
                        if (getSide() != piece.getSide()) {
                            capturables.add(piece);
                        }
                    }
                }
                if (getY() < 9) {
                    Piece piece = board.getPieceAt(getX() - 2, getY() + 1);
                    if (piece != null) {
                        if (getSide() != piece.getSide()) {
                            capturables.add(piece);
                        }
                    }
                }
            }
        }
        // 右邊。
        if (getX() < 7) {
            if (!board.isOccupiedAt(getX() + 1, getY())) {
                if (getY() > 0) {
                    Piece piece = board.getPieceAt(getX() + 2, getY() - 1);
                    if (piece != null) {
                        if (getSide() != piece.getSide()) {
                            capturables.add(piece);
                        }
                    }
                }
                if (getY() < 9) {
                    Piece piece = board.getPieceAt(getX() + 2, getY() + 1);
                    if (piece != null) {
                        if (getSide() != piece.getSide()) {
                            capturables.add(piece);
                        }
                    }
                }
            }
        }
        // 上面。
        if (getY() > 1) {
            if (!board.isOccupiedAt(getX(), getY() - 1)) {
                if (getX() > 0) {
                    Piece piece = board.getPieceAt(getX() - 1, getY() - 2);
                    if (piece != null) {
                        if (getSide() != piece.getSide()) {
                            capturables.add(piece);
                        }
                    }
                }
                if (getX() < 8) {
                    Piece piece = board.getPieceAt(getX() + 1, getY() - 2);
                    if (piece != null) {
                        if (getSide() != piece.getSide()) {
                            capturables.add(piece);
                        }
                    }
                }
            }
        }
        // 下面。
        if (getY() < 8) {
            if (!board.isOccupiedAt(getX(), getY() + 1)) {
                if (getX() > 0) {
                    Piece piece = board.getPieceAt(getX() - 1, getY() + 2);
                    if (piece != null) {
                        if (getSide() != piece.getSide()) {
                            capturables.add(piece);
                        }
                    }
                }
                if (getX() < 8) {
                    Piece piece = board.getPieceAt(getX() + 1, getY() + 2);
                    if (piece != null) {
                        if (getSide() != piece.getSide()) {
                            capturables.add(piece);
                        }
                    }
                }
            }
        }
        
        // 刪掉會被將軍，或無法解套將軍的吃法。
        removeCheckmatedEats(capturables);
        return capturables;
    }
    
    /**
     * 取得可以走的空格。
     * 
     * @return 可以走的空格。
     */
    public List<Point> getMovables() {
        Board board = (Board) getWorld();
        List<Point> movables = new ArrayList<Point>();
        
        // 左邊。
        if (getX() > 1) {
            if (!board.isOccupiedAt(getX() - 1, getY())) {
                if (getY() > 0) {
                    if (!board.isOccupiedAt(getX() - 2, getY() - 1)) {
                        movables.add(new Point(getX() - 2, getY() - 1));
                    }
                }
                if (getY() < 9) {
                    if (!board.isOccupiedAt(getX() - 2, getY() + 1)) {
                        movables.add(new Point(getX() - 2, getY() + 1));
                    }
                }
            }
        }
        // 右邊。
        if (getX() < 7) {
            if (!board.isOccupiedAt(getX() + 1, getY())) {
                if (getY() > 0) {
                    if (!board.isOccupiedAt(getX() + 2, getY() - 1)) {
                        movables.add(new Point(getX() + 2, getY() - 1));
                    }
                }
                if (getY() < 9) {
                    if (!board.isOccupiedAt(getX() + 2, getY() + 1)) {
                        movables.add(new Point(getX() + 2, getY() + 1));
                    }
                }
            }
        }
        // 上面。
        if (getY() > 1) {
            if (!board.isOccupiedAt(getX(), getY() - 1)) {
                if (getX() > 0) {
                    if (!board.isOccupiedAt(getX() - 1, getY() - 2)) {
                        movables.add(new Point(getX() - 1, getY() - 2));
                    }
                }
                if (getX() < 8) {
                    if (!board.isOccupiedAt(getX() + 1, getY() - 2)) {
                        movables.add(new Point(getX() + 1, getY() - 2));
                    }
                }
            }
        }
        // 下面。
        if (getY() < 8) {
            if (!board.isOccupiedAt(getX(), getY() + 1)) {
                if (getX() > 0) {
                    if (!board.isOccupiedAt(getX() - 1, getY() + 2)) {
                        movables.add(new Point(getX() - 1, getY() + 2));
                    }
                }
                if (getX() < 8) {
                    if (!board.isOccupiedAt(getX() + 1, getY() + 2)) {
                        movables.add(new Point(getX() + 1, getY() + 2));
                    }
                }
            }
        }
        
        // 刪掉會被將軍，或無法解套將軍的走法。
        removeCheckmatedMoves(movables);
        return movables;
    }
    
    /**
     * 另一格是不是在移動範圍內。
     * 
     * @param targetX 另一格的 X 座標
     * @param targetY 另一格的 Y 座標
     * @return 如果另一格在移動範圍內，就傳回 true ，否則傳回 false 。
     */
    private boolean isInMovingRange(int targetX, int targetY) {
        Board board = (Board) getWorld();
        
        // 日字形， 1x2 或 2x1 。
        int dx = targetX - getX();
        int dy = targetY - getY();
        if (dx * dx + dy * dy != 5) {
            return false;
        }
        
        // 向右拐馬腳。
        if (dx == 2) {
            if (board.isOccupiedAt(getX() + 1, getY())) {
                return false;
            }
        // 向左拐馬腳。
        } else if (dx == -2) {
            if (board.isOccupiedAt(getX() - 1, getY())) {
                return false;
            }
        // 向下拐馬腳。
        } else if (dy == 2) {
            if (board.isOccupiedAt(getX(), getY() + 1)) {
                return false;
            }
        // 向上拐馬腳。
        } else if (dy == -2) {
            if (board.isOccupiedAt(getX(), getY() - 1)) {
                return false;
            }
        }
        
        return true;
    }
}