import greenfoot.*;  // （World、Actor、GreenfootImage、GreenfootSound、Greenfoot及MouseInfo）
import java.awt.Point;
import java.util.ArrayList;
import java.util.List;

/**
 * 象棋棋子。
 * 
 * @author 依瑪貓
 * @version 2013-03-17
 */
public abstract class Piece extends Actor {
    
    /** 棋子所屬的陣營，紅方還是黑方。 */
    private ChessSide side = null;
    
    /** 棋子陣地是不是在棋盤下方。 */
    private boolean isBottomSide = true;
    
    /** 是不是第一次加進棋盤。 */
    private boolean isFirstAdded = true;
    
    /** 是不是正在拖拉中。避免重複進入移棋狀態。 */
    private boolean isDragging = false;
    
    /** 移棋起始位置的 X 座標值。 */
    private int fromX = -1;
    
    /** 移棋起始位置的 Y 座標值。 */
    private int fromY = -1;
    
    /**
     * 建立新的象棋棋子。
     * 
     * @param chessSide 棋子所屬的陣營，紅方或黑方。
     */
    public Piece(ChessSide chessSide) {
        side = chessSide;
        setImage(getClass().getName().toLowerCase() + "-" + chessSide + ".png");
    }
    
    /**
     * 棋子加進棋盤後，要執行的動作。
     * 
     * @param world 棋盤。
     */
    protected void addedToWorld(World world) {
        if (isFirstAdded) {
            Board board = (Board) world;
            
            if (board.isRedAtBottom() && side == ChessSide.BLACK) {
                GreenfootImage image = getImage();
                image.rotate(180);
            }
            if (!board.isRedAtBottom() && side == ChessSide.RED) {
                GreenfootImage image = getImage();
                image.rotate(180);
            }
            switch (side) {
            case RED:
                isBottomSide = board.isRedAtBottom();
                break;
            case BLACK:
                isBottomSide = !board.isRedAtBottom();
                break;
            }
            
            isFirstAdded = false;
        }
    }
    
    /**
     * 每一回合執行的動作。
     * 
     */
    public void act() {
        Board board = (Board) getWorld();
        
        if (side == board.getCurrentSide()) {
            // 開始拖拉棋子。標示可移動的位置。
            if (!isDragging && Greenfoot.mouseDragged(this)) {
                isDragging = true;
                fromX = getX();
                fromY = getY();
                tagAvailableMoves();
            }
            // 棋子拖拉中。只移動到可移動的位置，其餘不管它。
            if (Greenfoot.mouseDragged(this)) {
                MouseInfo mouse = Greenfoot.getMouseInfo();
                List<Hint> hints = board.getObjectsAt(mouse.getX(), mouse.getY(), Hint.class);
                if (hints.size() > 0) {
                    setLocation(mouse.getX(), mouse.getY());
                }
            }
            if (Greenfoot.mouseDragEnded(this)) {
                isDragging = false;
                List<Hint> hints = board.getObjects(Hint.class);
                board.removeObjects(hints);
                if (getX() != fromX || getY() != fromY) {
                    List<Piece> pieces = board.getObjectsAt(getX(), getY(), Piece.class);
                    pieces.remove(this);
                    board.removeObjects(pieces);
                    board.switchSide();
                    // 先 switchSide() 再歸零，因為 switchSide() 裏會加一。
                    if (pieces.size() > 0) {
                        board.resetLastCapture();
                    }
                }
            }
        }
    }
    
    /**
     * 取得棋子所屬的陣營。
     * 
     * @return 棋子所屬的陣營，紅方或黑方。
     */
    public ChessSide getSide() {
        return side;
    }
    
    /**
     * 棋子陣地是不是在棋盤下方。
     * 
     * @return 棋子陣地在棋盤下方時傳回 true ，否則傳回 false。
     */
    public boolean isBottomSide() {
        return isBottomSide;
    }
    
    /**
     * 能不能吃掉另一顆棋子。須在子類別中覆寫這個方法。
     * 
     * @param piece 另一顆棋子
     * @return 如果可以吃掉那顆棋子，就傳回 true ，否則傳回 false 。
     */
    public abstract boolean isCapturable(Piece another);
    
    /**
     * 能不能移到另一格。須在子類別中覆寫這個方法。
     * 
     * @param x 另一格的 X 座標a
     * @param y 另一格的 Y 座標
     * @return 如果可以移到那一格，就傳回 true ，否則傳回 false 。
     */
    public abstract boolean isMovable(int x, int y);
    
    /**
     * 取得可以吃的棋子。須在子類別中覆寫這個方法。
     * 
     * @return 可以吃的棋子。
     */
    public abstract List<Piece> getCapturables();
    
    /**
     * 取得可以走的空格。須在子類別中覆寫這個方法。
     * 
     * @return 可以走的空格。
     */
    public abstract List<Point> getMovables();
    
    
    /**
     * 棋子所在是不是符合規則的位置上。預設是 true （俥／車、傌／馬、炮／包的位置沒有規則）。
     * 
     * @return 如果棋子在符合規則的位置上，就傳回 true ，否則傳回 false 。
     */
    public boolean isAtLegalPosition() {
        return true;
    }
    
    /**
     * 標示下一步可以走的位置。
     * 
     */
    public void tagAvailableMoves() {
        Board board = (Board) getWorld();
        int x = getX(), y = getY();
        
        board.addObject(new Hint(), x, y);
        for (Piece target : getCapturables()) {
            board.addObject(new Hint(), target.getX(), target.getY());
        }
        for (Point target : getMovables()) {
            board.addObject(new Hint(), target.x, target.y);
        }
        // 重新加入自己，把自己疊在標示的上方。
        board.removeObject(this);
        board.addObject(this, x, y);
    }
    
    /**
     * 刪掉會被將軍，或無法解套將軍的吃法。
     * 
     * @param targets 下一步要吃的目標
     */
    public void removeCheckmatedEats(List<Piece> targets) {
        Board board = (Board) getWorld();
        List<Piece> candidates = new ArrayList<Piece>(targets);
        General general = getMyGeneral();
        int x = getX(), y = getY();
        
        if (general == null) {
            targets.clear();
            return;
        }
        
        for (Piece target : candidates) {
            int targetX = target.getX();
            int targetY = target.getY();
            boolean isCheckmated = false;
            
            board.removeObject(target);
            setLocation(targetX, targetY);
            isCheckmated = general.isCheckmated();
            board.addObject(target, targetX, targetY);
            setLocation(x, y);
            if (isCheckmated) {
                targets.remove(target);
            }
        }
    }
    
    /**
     * 刪掉會被將軍，或無法解套將軍的走法。
     * 
     * @param targets 下一步要走的目標
     */
    public void removeCheckmatedMoves(List<Point> targets) {
        Board board = (Board) getWorld();
        List<Point> candidates = new ArrayList<Point>(targets);
        General general = getMyGeneral();
        int x = getX(), y = getY();
        
        if (general == null) {
            targets.clear();
            return;
        }
        
        for (Point target : candidates) {
            boolean isCheckmated = false;
            
            setLocation(target.x, target.y);
            isCheckmated = general.isCheckmated();
            setLocation(x, y);
            if (isCheckmated) {
                targets.remove(target);
            }
        }
    }
    
    /**
     * 取得我方的帥／將。
     * 
     * @return 我方的帥／將
     */
    private General getMyGeneral() {
        Board board = (Board) getWorld();
        List<General> generals = board.getObjects(General.class);
        for (General general : generals) {
            if (general.getSide() == side) {
                return general;
            }
        }
        // 不可能走到這裏。
        return null;
    }
}