import greenfoot.*;  // （World、Actor、GreenfootImage、GreenfootSound、Greenfoot及MouseInfo）
import java.awt.Point;
import java.util.ArrayList;
import java.util.List;

/**
 * 象棋的兵／卒。
 * 
 * @author 依瑪貓
 * @version 2013-03-17
 */
public class Soldier extends Piece {
    
    /**
     * 建立新的兵／卒。
     * 
     * @param chessSide 棋子所屬的陣營，紅方或黑方。
     */
    public Soldier(ChessSide chessSide) {
        super(chessSide);
    }
    
    /**
     * 能不能吃掉另一顆棋子。
     * 
     * @param piece 另一顆棋子
     * @return 如果可以吃掉那顆棋子，就傳回 true ，否則傳回 false 。
     */
    public boolean isCapturable(Piece another) {
        // 棋子只能放在符合規則的位置上。
        if (!isAtLegalPosition()) {
            return false;
        }
        // 不能吃同一方的棋子。
        if (getSide() == another.getSide()) {
            return false;
        }
        // 檢查另一顆棋子是不是在移動範圍內。
        return isInMovingRange(another.getX(), another.getY());
    }
    
    /**
     * 能不能移到另一格。
     * 
     * @param targetX 另一格的 X 座標
     * @param targetY 另一格的 Y 座標
     * @return 如果可以移到那一格，就傳回 true ，否則傳回 false 。
     */
    public boolean isMovable(int targetX, int targetY) {
        Board board = (Board) getWorld();
        
        // 棋子只能放在符合規則的位置上。
        if (!isAtLegalPosition()) {
            return false;
        }
        // 只能移到空格。
        if (board.isOccupiedAt(targetX, targetY)) {
            return false;
        }
        // 檢查目標格是不是在移動範圍內。
        return isInMovingRange(targetX, targetY);
    }
    
    /**
     * 取得可以吃的棋子。
     * 
     * @return 可以吃的棋子。
     */
    public List<Piece> getCapturables() {
        Board board = (Board) getWorld();
        List<Piece> capturables = new ArrayList<Piece>();
        
        // 棋子只能放在符合規則的位置上。
        if (!isAtLegalPosition()) {
            return capturables;
        }
        
        if (isBottomSide()) {
            // 未過河。
            if (getY() > 4) {
                Piece piece = board.getPieceAt(getX(), getY() - 1);
                if (piece != null) {
                    if (getSide() != piece.getSide()) {
                        capturables.add(piece);
                    }
                }
            
            // 已過河。
            } else {
                Piece piece = null;
                piece = board.getPieceAt(getX(), getY() - 1);
                if (piece != null) {
                    if (getSide() != piece.getSide()) {
                        capturables.add(piece);
                    }
                }
                piece = board.getPieceAt(getX() - 1, getY());
                if (piece != null) {
                    if (getSide() != piece.getSide()) {
                        capturables.add(piece);
                    }
                }
                piece = board.getPieceAt(getX() + 1, getY());
                if (piece != null) {
                    if (getSide() != piece.getSide()) {
                        capturables.add(piece);
                    }
                }
            }
        
        } else {
            // 未過河。
            if (getY() < 5) {
                Piece piece = board.getPieceAt(getX(), getY() + 1);
                if (piece != null) {
                    if (getSide() != piece.getSide()) {
                        capturables.add(piece);
                    }
                }
            
            // 已過河。
            } else {
                Piece piece = null;
                piece = board.getPieceAt(getX(), getY() + 1);
                if (piece != null) {
                    if (getSide() != piece.getSide()) {
                        capturables.add(piece);
                    }
                }
                piece = board.getPieceAt(getX() - 1, getY());
                if (piece != null) {
                    if (getSide() != piece.getSide()) {
                        capturables.add(piece);
                    }
                }
                piece = board.getPieceAt(getX() + 1, getY());
                if (piece != null) {
                    if (getSide() != piece.getSide()) {
                        capturables.add(piece);
                    }
                }
            }
        }
        
        // 刪掉會被將軍，或無法解套將軍的吃法。
        removeCheckmatedEats(capturables);
        return capturables;
    }
    
    
    /**
     * 取得可以走的空格。
     * 
     * @return 可以走的空格。
     */
    public List<Point> getMovables() {
        Board board = (Board) getWorld();
        List<Point> movables = new ArrayList<Point>();
        
        // 棋子只能放在符合規則的位置上。
        if (!isAtLegalPosition()) {
            return movables;
        }
        
        if (isBottomSide()) {
            // 未過河。
            if (getY() > 4) {
                if (!board.isOccupiedAt(getX(), getY() - 1)) {
                    movables.add(new Point(getX(), getY() - 1));
                }
            
            // 已過河。
            } else {
                if (!board.isOccupiedAt(getX(), getY() - 1)) {
                    movables.add(new Point(getX(), getY() - 1));
                }
                if (!board.isOccupiedAt(getX() - 1, getY())) {
                    movables.add(new Point(getX() - 1, getY()));
                }
                if (!board.isOccupiedAt(getX() + 1, getY())) {
                    movables.add(new Point(getX() + 1, getY()));
                }
            }
        
        } else {
            // 未過河。
            if (getY() < 5) {
                if (!board.isOccupiedAt(getX(), getY() + 1)) {
                    movables.add(new Point(getX(), getY() + 1));
                }
            
            // 已過河。
            } else {
                if (!board.isOccupiedAt(getX(), getY() + 1)) {
                    movables.add(new Point(getX(), getY() + 1));
                }
                if (!board.isOccupiedAt(getX() - 1, getY())) {
                    movables.add(new Point(getX() - 1, getY()));
                }
                if (!board.isOccupiedAt(getX() + 1, getY())) {
                    movables.add(new Point(getX() + 1, getY()));
                }
            }
        }
        
        // 刪掉會被將軍，或無法解套將軍的走法。
        removeCheckmatedMoves(movables);
        return movables;
    }
    
    /**
     * 棋子所在是不是符合規則的位置上。
     * 
     * @return 如果棋子在符合規則的位置上，就傳回 true ，否則傳回 false 。
     */
    public boolean isAtLegalPosition() {
        int y0 = -1, y1 = -1, dy = 0;
        
        if (isBottomSide()) {
            // 未過河。
            if (getY() > 4) {
                if (getX() % 2 == 1 || getY() > 6) {
                    return false;
                }
                return true;
            }
        
        } else {
            // 未過河。
            if (getY() < 5) {
                if (getX() % 2 == 1 || getY() < 3) {
                    return false;
                }
                return true;
            }
        }
        return true;
    }
    
    /**
     * 另一格是不是在移動範圍內。
     * 
     * @param targetX 另一格的 X 座標
     * @param targetY 另一格的 Y 座標
     * @return 如果另一格在移動範圍內，就傳回 true ，否則傳回 false 。
     */
    private boolean isInMovingRange(int targetX, int targetY) {
        int dx = targetX - getX();
        int dy = targetY - getY();
        
        if (isBottomSide()) {
            if (getY() > 4) {
                if (dx == 0 && dy == -1) {
                    return true;
                }
                return false;
            } else {
                if (dx * dx + dy * dy == 1 && dy <= 0) {
                    return true;
                }
                return false;
            }
        
        } else {
            if (getY() < 5) {
                if (dx == 0 && dy == 1) {
                    return true;
                }
                return false;
            } else {
                if (dx * dx + dy * dy == 1 && dy >= 0) {
                    return true;
                }
                return false;
            }
        }
    }
}